pub mod fileupload {

    pub use actix_multipart::form::{tempfile::TempFile, MultipartForm};
    pub use actix_web::{get, post, web, App, HttpResponse, HttpServer, HttpRequest, Responder};
    pub use std::fs;
    pub use std::io::{Read, Write};
    pub use serde_derive::{Serialize, Deserialize};
    pub use base64::prelude::*;
    pub use actix_files::NamedFile;
    pub use rand::Rng;
    pub use chrono::Local;
    pub use magic_crypt::{new_magic_crypt, MagicCryptTrait};
    pub use std::fs::OpenOptions;
    pub use humansize::{format_size, DECIMAL};
    pub use actix_web::Error;
    pub use actix_multipart::MultipartError;



    #[derive(Debug, MultipartForm)]
    pub struct UploadForm {
        #[multipart(limit = "100MB")]
        file: TempFile,
    }

    #[derive(Debug, Serialize, Deserialize, Clone)]
    pub struct EncryptedFile {
        name: String,
        crypt: String,
        date: String,
        id: u64,
    }


    fn b64_to_file(b64: &String, destfile: &str) -> Result<(), Box<dyn std::error::Error>> {
        let decoded_data = BASE64_STANDARD.decode(b64)?;
        fs::write(destfile, &decoded_data)?;
        Ok(())
    }

    fn encrypt(content: &String, password: &String) -> String {
        let mc = new_magic_crypt!(password, 256);
        let encrypted = mc.encrypt_str_to_base64(&content);
        return String::from(encrypted);
    }

    fn random_password() -> String {
        const CHARSET: &[u8] = b"ABCDEFGHIJKLMNOPQRSTUVWXYZ\
                                abcdefghijklmnopqrstuvwxyz\
                                0123456789";
        let mut rng = rand::thread_rng();
        let password: String = (0..64)
            .map(|_| {
                let idx = rng.gen_range(0..CHARSET.len());
                CHARSET[idx] as char
            })
            .collect();
        return password;
    }

    fn store_encrypted_file(b64: &String, name: &String, password: &String, id: &u64) -> Result<(), Box<dyn std::error::Error>> {
        let mut file = OpenOptions::new()
            .write(true)
            .append(true)
            .create(true)
            .open(format!("./Files/Stored_files/{}.json", id))
            .unwrap();

        let crypt = encrypt(b64, password);
        let now = Local::now();
        let now = now.format("%d/%m/%Y@%H:%M:%S");
        let my_encrypted_file = EncryptedFile { name: name.clone(), crypt: crypt, date: format!("{}", now), id: id.clone() };
        let json = serde_json::to_string(&my_encrypted_file).unwrap();

        writeln!(file, "{}", &json)?;
        Ok(())

    }

    fn get_decrypted_file(id: &u64, password: &String) ->  Result<String, Box<dyn std::error::Error>> {
        let json_content = fs::read_to_string(format!("./Files/Stored_files/{}.json", id))?;
        let encr_file: EncryptedFile = serde_json::from_str(json_content.as_str())?;
        let decr_file = decrypt(encr_file.crypt.clone(), password.clone().as_str())?;
        let path = format!("/tmp/{}", encr_file.name);
        b64_to_file(&decr_file, &path)?;
        Ok(path)
    }
     
        
    fn decrypt(encrypted: String, pass: &str) -> Result<String, String> {
        let mc = new_magic_crypt!(pass, 256);
        let result = mc.decrypt_base64_to_string(&encrypted);
        match result {
            Ok(v) => {
                return Ok(v);
            },
            Err(_) => {
                return Err("Could not decrypt".to_string());
            },
        }
    }

    #[post("/file")]
    pub async fn post_file(MultipartForm(form): MultipartForm<UploadForm>, req: HttpRequest) -> impl Responder {

        //File is :
        //UploadForm { file: TempFile { file: NamedTempFile("/tmp/.tmp8yLb6N"), content_type: Some("image/jpeg"), file_name: Some("157.jpg"), size: 1437187 } }

        //The file is written to /tmp, which means trying to persist it in current dir fails.
        //But we can read it directly instead.


        //Persisting == writing the temporary file to a definitive file.
        //let destpath = String::from("/tmp/destfile.jpg");
        //let stored_file = form.file.file.persist(destpath).unwrap();

        

        //Reading the file in base64.
        let mut src_file = form.file.file.reopen().unwrap();
        let mut buffer = Vec::new();
        src_file.read_to_end(&mut buffer).unwrap();
        let base64_content = BASE64_STANDARD.encode(&buffer);

        //Writing it back on disk.
        //let destpath = String::from("./coucou.jpg");
        //b64_to_file(&base64_content, &destpath).unwrap();

        //Storing it encrypted.
        let password = random_password();
        let mut rng = rand::thread_rng();
        let id: u64 = rng.gen::<u64>();
        let hostname = req.headers().get("Host").unwrap().to_str().unwrap();
        let filename = String::from(form.file.file_name.unwrap_or("NAMENOTFOUND".to_string()));
        match store_encrypted_file(&base64_content, &filename, &password, &id){
            Ok(_) => (),
            Err(e) => return HttpResponse::Ok().body(format!("File could not be saved, got {}", e)),
        };

        //try to get html header and footer files.
        let header = fs::read_to_string("./Files/header.html").unwrap_or("".to_string());
        let footer = fs::read_to_string("./Files/footer.html").unwrap_or("".to_string());
        let hsize = format_size(form.file.size, DECIMAL);

        //We have to send smthing back to the user.
        let link = format!("http://{}/getfile/{}/{}", hostname, id, password); 
        let answer = format!(
            "{}Uploaded {} with size: {}\n\
            <br><br><a href=\"{}\">Here</a>\n\
            <br><br>You can download it with in a terminal with<br>\
            <pre>curl -JLO {}</pre>{}",
            header,
            filename,
            hsize,
            link,
            link,
            footer,
        );
        return HttpResponse::Ok().body(answer);
    }

    //Serve the file back to the user
    #[get("/getfile/{id}/{password}")]
    pub async fn get_file(path: web::Path<(u64, String)>) -> impl Responder {
        let (id,password) = path.into_inner();
        let path = get_decrypted_file(&id, &password).unwrap();
        let sendback = NamedFile::open(path.clone());
        match fs::remove_file(path.clone()) {
            Ok(_) => (),
            Err(e) => eprintln!("File {} could not be deleted, got {}", path, e),
        };
        sendback
    }


    #[get("/")]
    pub async fn get_index() -> impl Responder {
        let page = fs::read_to_string("./Files/index.html").unwrap();
        return HttpResponse::Ok().body(page);
    }

    pub fn handle_multipart_error(err: MultipartError, _req: &HttpRequest) -> Error {
        eprintln!("Multipart error {}", err);
        return Error::from(err);
    }
}
